"""
Tests for specific issues and pull requests
"""


import os
import tempfile
import difflib
from pathlib import Path
from textwrap import dedent
import gffutils
from gffutils import feature
from gffutils import helpers
from gffutils.gffwriter import GFFWriter
from gffutils import merge_criteria as mc

import pytest


def test_issue_79():
    gtf = gffutils.example_filename("keep-order-test.gtf")
    db = gffutils.create_db(
        gtf,
        "tmp.db",
        disable_infer_genes=False,
        disable_infer_transcripts=False,
        id_spec={"gene": "gene_id", "transcript": "transcript_id"},
        merge_strategy="create_unique",
        keep_order=True,
        force=True,
    )

    exp = open(gtf).read()
    obs = "\n".join([str(i) for i in db.all_features()])
    exp_1 = exp.splitlines(True)[0].strip()
    obs_1 = obs.splitlines(True)[0].strip()
    print("EXP")
    print(exp_1)
    print("OBS")
    print(obs_1)
    print("DIFF")
    print("".join(difflib.ndiff([exp_1], [obs_1])))
    assert obs_1 == exp_1


def test_issue_82():
    # key-val separator is inside an unquoted attribute value
    x = (
        "Spenn-ch12\tsgn_markers\tmatch\t2621812\t2622049\t.\t+\t.\t"
        "Alias=SGN-M1347;ID=T0028;Note=marker name(s): T0028 SGN-M1347 |identity=99.58|escore=2e-126"
    )
    y = feature.feature_from_line(x)
    assert y.attributes["Note"] == [
        "marker name(s): T0028 SGN-M1347 |identity=99.58|escore=2e-126"
    ]

    gffutils.create_db(gffutils.example_filename("keyval_sep_in_attrs.gff"), ":memory:")


def test_issue_85():
    # when start or stop was empty, #85 would fail Should now work with
    # blank fields
    f = feature.feature_from_line("\t".join([""] * 9))

    # or with "." placeholders
    f = feature.feature_from_line("\t".join(["."] * 9))


def test_issue_105():
    fn = gffutils.example_filename("FBgn0031208.gtf")
    home = os.path.expanduser("~")
    newfn = os.path.join(home, ".gffutils.test")
    with open(newfn, "w") as fout:
        fout.write(open(fn).read())
    f = gffutils.iterators.DataIterator(newfn)
    for i in f:
        pass
    os.unlink(newfn)


def test_issue_107():
    s = dedent(
        """
        chr1\t.\tgene\t10\t15\t.\t+\t.\tID=b;
        chr1\t.\tgene\t1\t5\t.\t-\t.\tID=a;
        chr2\t.\tgene\t25\t50\t.\t-\t.\tID=c;
        chr2\t.\tgene\t55\t60\t.\t-\t.\tID=d;
        """
    )
    tmp = tempfile.NamedTemporaryFile(delete=False).name
    with open(tmp, "w") as fout:
        fout.write(s + "\n")
    db = gffutils.create_db(tmp, ":memory:")
    interfeatures = list(
        db.interfeatures(db.features_of_type("gene", order_by=("seqid", "start")))
    )
    assert [str(i) for i in interfeatures] == [
        "chr1\tgffutils_derived\tinter_gene_gene\t6\t9\t.\t.\t.\tID=a-b;",
        "chr2\tgffutils_derived\tinter_gene_gene\t51\t54\t.\t-\t.\tID=c-d;",
    ]


def test_issue_119():
    # First file has these two exons with no ID:
    #
    #   chr2L FlyBase exon  8193  8589  .  +  .  Parent=FBtr0300690
    #   chr2L FlyBase exon  7529  8116  .  +  .  Name=CG11023:1;Parent=FBtr0300689,FBtr0300690
    #
    db0 = gffutils.create_db(gffutils.example_filename("FBgn0031208.gff"), ":memory:")

    # And this one, a bunch of reads with no IDs anywhere
    db1 = gffutils.create_db(
        gffutils.example_filename("F3-unique-3.v2.gff"), ":memory:"
    )

    # When db1 is updated by db0
    db2 = db1.update(db0)
    assert (
        db2._autoincrements == db1._autoincrements == {"exon": 2, "read": 112}
    ), db2._autoincrements

    assert len(list(db0.features_of_type("exon"))) == 6

    # Now we update that with db0 again
    db3 = db2.update(db0, merge_strategy="replace")

    # Using the "replace" strategy, we should have only gotten another 2 exons
    assert len(list(db3.features_of_type("exon"))) == 8

    # Make sure that the autoincrements for exons jumped by 2
    assert (
        db2._autoincrements == db3._autoincrements == {"exon": 4, "read": 112}
    ), db2._autoincrements

    # More isolated test, merging two databases each created from the same file
    # which itself contains only a single feature with no ID.
    tmp = tempfile.NamedTemporaryFile(delete=False).name
    with open(tmp, "w") as fout:
        fout.write("chr1\t.\tgene\t10\t15\t.\t+\t.\t\n")

    db4 = gffutils.create_db(tmp, tmp + ".db")
    db5 = gffutils.create_db(tmp, ":memory:")

    assert db4._autoincrements == {"gene": 1}
    assert db5._autoincrements == {"gene": 1}

    db6 = db4.update(db5)

    db7 = gffutils.FeatureDB(db4.dbfn)

    # both db4 and db6 should now have the same, updated autoincrements because
    # they both point to the same db.
    assert db6._autoincrements == db4._autoincrements == {"gene": 2}

    # But db5 was created independently and should have unchanged autoincrements
    assert db5._autoincrements == {"gene": 1}

    # db7 was created from the database pointed to by both db4 and db6. This
    # tests that when a FeatureDB is created it should have the
    # correctly-updated autoincrements read from the db
    assert db7._autoincrements == {"gene": 2}


def test_pr_131():
    db = gffutils.create_db(gffutils.example_filename("FBgn0031208.gff"), ":memory:")

    # previously would raise ValueError("No lines parsed -- was an empty
    # file provided?"); now just does nothing
    db2 = db.update([])


def test_pr_133():
    # Previously, merge_attributes would not deep-copy the values from the
    # second dict, and when the values are then modified, the second dict is
    # unintentionally modified.
    d1 = {"a": [1]}
    d2 = {"a": [2]}
    d1a = {"a": [1]}
    d2a = {"a": [2]}
    d3 = gffutils.helpers.merge_attributes(d1, d2)
    assert d1 == d1a, d1
    assert d2 == d2a, d2


def test_pr_139():
    db = gffutils.create_db(gffutils.example_filename("FBgn0031208.gff"), ":memory:")
    exons = list(db.features_of_type("exon"))
    inter = list(db.interfeatures(exons))

    # previously, the first exon's attributes would show up in subsequent merged features
    first_name = exons[0].attributes["Name"][0]
    for i in inter[1:]:
        if "Name" in i.attributes:
            assert first_name not in i.attributes["Name"], str(i)


def test_pr_144():
    # previously this would fail with:
    #   UnboundLocalError: local variable 'part' referenced before assignment
    f = gffutils.Feature(attributes={"a": [""]})

    # Make sure everything got converted correctly
    assert f.attributes["a"] == [""]
    assert str(f) == ".	.	.	.	.	.	.	.	a"
    g = gffutils.feature.feature_from_line(str(f))
    g.dialect["fmt"] = "gff3"
    print(g.attributes)
    print(g.dialect)
    assert str(g) == str(f)


def test_pr_172():
    line = (
        "NC_049222.1\tGnomon\tgene\t209085\t282880\t.\t-\t.\t"
        'gene_id "ENPP1_3"; transcript_id ""; db_xref "GeneID:100856150";'
        'db_xref "VGNC:VGNC:40374"; gbkey "Gene"; gene "ENPP1"; '
        'gene_biotype "protein_coding";\n'
    )
    tmp = tempfile.NamedTemporaryFile(delete=False).name
    with open(tmp, "w") as fout:
        fout.write(line)
    db = gffutils.create_db(tmp, ":memory:")


def test_pr_171():
    q = gffutils.parser.Quoter()
    assert q.__missing__("\n") == "%0A"
    assert q.__missing__("a") == "a"

    assert q.__missing__("") == ""


def test_issue_129():

    # thanks @Brunox13 for the detailed notes on #129

    line = 'chr1\tdemo\tstart_codon\t69091\t69093\t.\t+\t.\tgene_id "demo";\n'
    tmp = tempfile.NamedTemporaryFile(delete=False).name
    with open(tmp, "w") as fout:
        fout.write(line)
    db = gffutils.create_db(tmp, ":memory:")

    # ASCII art to visualize each test (coords are along the top, from 69087 to
    # 69090). The tests slide a 4-bp region over the original 3-bp start codon.

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    # | | | |               Test feature
    res = list(db.region(region=("chr1", 69087, 69090), featuretype="start_codon"))
    assert len(res) == 0

    # NOTE: prior to #162, this did not return anything
    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #   | | | |             Test feature
    res = list(db.region(region=("chr1", 69088, 69091), featuretype="start_codon"))
    assert len(res) == 1

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #     | | | |           Test feature
    res = list(db.region(region=("chr1", 69089, 69092), featuretype="start_codon"))
    assert len(res) == 1

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #       | | | |         Test feature
    res = list(db.region(region=("chr1", 69090, 69093), featuretype="start_codon"))
    assert len(res) == 1

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #         | | | |       Test feature
    res = list(db.region(region=("chr1", 69091, 69094), featuretype="start_codon"))
    assert len(res) == 1

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #           | | | |     Test feature
    res = list(db.region(region=("chr1", 69092, 69095), featuretype="start_codon"))
    assert len(res) == 1

    # NOTE: priro to #162, this did not return anything
    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #             | | | |   Test feature
    res = list(db.region(region=("chr1", 69093, 69096), featuretype="start_codon"))
    assert len(res) == 1

    # 7 8 9 0 1 2 3 4 5 6 7
    #         | | |         Orig feature
    #               | | | | Test feature
    res = list(db.region(region=("chr1", 69094, 69097), featuretype="start_codon"))
    assert len(res) == 0


def test_issue_128():
    # In #128, some lines had separators of "; " and some with ";". The first
    # one in the file would win. Now the detection pays more attention to lines
    # with more attributes to make it work properly
    gff = gffutils.example_filename("gms2_example.gff3")
    db = gffutils.create_db(gff, ":memory:", force=True)
    expected = {
        "ID": ["1"],
        "Parent": ["gene_1"],
        "gene_type": ["native"],
        "partial": ["11"],
        "gc": ["33"],
        "length": ["363"],
    }
    assert dict(db["1"].attributes) == expected


def test_issue_157():
    # With the merge overhaul, children_bp incorrectly still used ignore_strand.
    db = gffutils.create_db(gffutils.example_filename("FBgn0031208.gff"), ":memory:")
    gene = next(db.features_of_type("gene"))
    children = list(db.children(gene, featuretype="exon"))

    # Modify the last one to have a different strand so we can test the
    # ignore_strand argument.
    children[-1].strand = "-"
    db.update(children[-1:], merge_strategy="replace")

    # and, since updating has been problematic in the past, double-check again
    # that the strand is changed in the db.
    assert list(db.children(gene, featuretype="exon"))[-1].strand == "-"
    cbp1 = db.children_bp(gene, child_featuretype="exon")

    # Previously this would give:
    #   TypeError: merge() got an unexpected keyword argument 'ignore_strand'
    #
    # Now changing to ValueError and suggesting a fix.
    with pytest.raises(ValueError):
        db.children_bp(gene, child_featuretype="exon", merge=True, ignore_strand=True)
    with pytest.raises(ValueError):
        db.children_bp(gene, ignore_strand=True, nonexistent=True)
    with pytest.raises(TypeError):
        db.children_bp(gene, nonexistent=True)

    # The way to do it now is the following (we can omit the mc.feature_type
    # since we're preselecting for exons anyway):
    db.children_bp(
        gene,
        child_featuretype="exon",
        merge=True,
        merge_criteria=(mc.overlap_end_inclusive),
    )


def test_issue_159():
    db = gffutils.create_db(gffutils.example_filename("FBgn0031208.gff"), ":memory:")
    fasta = gffutils.example_filename("dm6-chr2L.fa")
    for transcript, seq in gffutils.helpers.canonical_transcripts(db, fasta):
        pass


def test_issue_164():
    # Something strange with the original annotation, but seems fine at least
    # after pasting in the offending genes from the GitHub comments.
    db = gffutils.create_db(
        gffutils.example_filename("sharr.gtf"),
        ":memory:",
        disable_infer_transcripts=True,
        disable_infer_genes=True,
        id_spec={"gene": "gene_id", "transcript": "transcript_id"},
        merge_strategy="create_unique",
        keep_order=True,
    )


def test_issue_166():
    # Added the new FeatureDB.seqids() method.
    db = gffutils.create_db(gffutils.example_filename("nonascii"), ":memory:")
    seqs = list(db.seqids())
    assert seqs == ["2L", "2R", "3L", "3R", "X"], seqs


def test_issue_167():
    # Previously was causing sqlite3.InterfaceError
    db = gffutils.create_db(gffutils.example_filename("issue167.gff"), ":memory:")


def test_issue_174():
    db = gffutils.create_db(
        gffutils.example_filename("issue174.gtf"),
        ":memory:",
        merge_strategy="warning",
    )
    introns = [f for f in db.create_introns()]
    observed = [i.attributes["exon_number"] for i in introns]
    assert observed[7] == ["8", "9"]
    assert observed[8] == ["10", "9"]
    assert observed[9] == ["10", "11"]

    # Now do the same thing, but with the new numeric_sort arg
    introns = [f for f in db.create_introns(numeric_sort=True)]
    observed = [i.attributes["exon_number"] for i in introns]
    assert observed[7] == ["8", "9"]
    # This should be fixed:
    assert observed[8] == ["9", "10"]
    assert observed[9] == ["10", "11"]


def test_issue_197():

    # Previously this would fail with ValueError due to using the stop position
    # of the last item on the previous chrom as the start position.

    db = gffutils.create_db(
        gffutils.example_filename("issue_197.gff"), ":memory:", merge_strategy="error"
    )
    genes = list(db.features_of_type("gene"))
    igss = list(db.interfeatures(genes, new_featuretype="intergenic_space"))

    # Prior to PR #219, multiple IDs could be created by interfeatures, which
    # in turn was patched here by providing the transform to db.update. With
    # #219, this ends up being a no-op because ID is a single value by the time
    # it gets to the transform function.
    #
    # However, keeping the test as-is to ensure backward-compatibility.
    def transform(f):
        f["ID"] = ["-".join(f.attributes["ID"])]
        return f

    db = db.update(igss, transform=transform, merge_strategy="error")

    obs = list(db.features_of_type("intergenic_space"))
    for i in obs:
        print(i)

    assert [str(i) for i in obs] == [
        "tig00000492\tgffutils_derived\tintergenic_space\t47236\t47350\t.\t-\t.\tID=ctg492.gene0001-ctg492.gene0002;Name=gene0001,gene0002",
        "tig00000492\tgffutils_derived\tintergenic_space\t48257\t49999\t.\t-\t.\tID=ctg492.gene0002-gene0;Name=gene0002",
        "tig00000492\tgffutils_derived\tintergenic_space\t50050\t50054\t.\t-\t.\tID=gene3-gene4",
        "tig00000492\tgffutils_derived\tintergenic_space\t50071\t50071\t.\t-\t.\tID=gene4-gene5",
        "tig00000492\tgffutils_derived\tintergenic_space\t50076\t50089\t.\t-\t.\tID=gene5-gene6",
    ]


def test_issue_198():
    line = 'NC_000001.11	BestRefSeq	gene	14362	29370	.	-	.	gene_id "WASH7P"; transcript_id ""; db_xref "GeneID:653635"; db_xref "HGNC:HGNC:38034"; description "WASP family homolog 7, pseudogene"; gbkey "Gene"; gene "WASH7P"; gene_biotype "transcribed_pseudogene"; gene_synonym "FAM39F"; gene_synonym "WASH5P"; pseudo "true";'

    # Original issue #198 is that this fails with:
    #
    #   gffutils.exceptions.AttributeStringError: Internally inconsistent
    #   attributes formatting: some have repeated keys, some do not.
    #
    # This is because the dialect inference sees the two db_xref keys, and
    # correctly assumes the dialect uses repeated keys rather than
    # multiple, comma-separated values -- but there's a comma in the
    # description.
    #
    # So we need to figure out the best way of interpreting a comma in cases
    # like this. It seems like the best solution is to assume that the presence
    # of repeated keys always wins.
    f = feature.feature_from_line(line)

    assert f.attributes["description"] == ["WASP family homolog 7, pseudogene"]

    # If we remove one of the db_xref keys, then previously the parser saw the
    # comma and figured it was a multivalue key, and split it. Now, it's
    # correctly identified as a single-value key.
    #
    # Note that we still have gene_synonym as a repeated key.
    line = 'NC_000001.11	BestRefSeq	gene	14362	29370	.	-	.	gene_id "WASH7P"; transcript_id ""; db_xref "GeneID:653635"; description "WASP family homolog 7, pseudogene"; gbkey "Gene"; gene "WASH7P"; gene_biotype "transcribed_pseudogene"; gene_synonym "FAM39F"; gene_synonym "WASH5P"; pseudo "true";'
    f = feature.feature_from_line(line)
    assert f.dialect["repeated keys"]

    # Previous result, note leading space --------------------------->| |
    # assert f.attributes['description'] == ['WASP family homolog 7', ' pseudogene']

    # Current result: not split.
    assert f.attributes["description"] == ["WASP family homolog 7, pseudogene"]

    # Removing that space before "pseudogene" might mean it's a multivalue, but
    # we decide on the convention that if keys are repeated at all, that wins.
    # So we still don't split
    line = 'NC_000001.11	BestRefSeq	gene	14362	29370	.	-	.	gene_id "WASH7P"; transcript_id ""; db_xref "GeneID:653635"; description "WASP family homolog 7,pseudogene"; gbkey "Gene"; gene "WASH7P"; gene_biotype "transcribed_pseudogene"; gene_synonym "FAM39F"; gene_synonym "WASH5P"; pseudo "true";'
    f = feature.feature_from_line(line)
    assert f.dialect["repeated keys"]
    assert f.attributes["description"] == ["WASP family homolog 7,pseudogene"]

    # But if we get rid of all repeated keys, it's interpreted as multiple values
    line = 'NC_000001.11	BestRefSeq	gene	14362	29370	.	-	.	gene_id "WASH7P"; transcript_id ""; db_xref "GeneID:653635"; description "WASP family homolog 7,pseudogene"; gbkey "Gene"; gene "WASH7P"; gene_biotype "transcribed_pseudogene"; gene_synonym "FAM39F"; pseudo "true";'
    f = feature.feature_from_line(line)
    assert not f.dialect["repeated keys"]
    assert f.attributes["description"] == ["WASP family homolog 7", "pseudogene"]

    # ....but if there's a ", " (comma followed by space) instead of just
    # comma, then it's not split.
    line = 'NC_000001.11	BestRefSeq	gene	14362	29370	.	-	.	gene_id "WASH7P"; transcript_id ""; db_xref "GeneID:653635"; description "WASP family homolog 7, pseudogene"; gbkey "Gene"; gene "WASH7P"; gene_biotype "transcribed_pseudogene"; gene_synonym "FAM39F"; pseudo "true";'
    f = feature.feature_from_line(line)
    assert not f.dialect["repeated keys"]
    assert f.attributes["description"] == ["WASP family homolog 7, pseudogene"]


    # Confirm behavior of corner cases like a trailing comma
    line = "chr17	RefSeq	CDS	6806527	6806553	.	+	0	Name=CDS:NC_000083.5:LOC100040603;Parent=XM_001475631.1,"
    f = feature.feature_from_line(line)
    assert f.attributes["Parent"] == ["XM_001475631.1", ""]


def test_issue_207():
    def _check(txt, expected_keys, dialect_trailing_semicolon):
        db = gffutils.create_db(txt.replace(" ", "\t"), ":memory:", from_string=True)
        assert [list(f.attributes.keys()) for f in db.all_features()] == expected_keys
        assert db.dialect["trailing semicolon"] == dialect_trailing_semicolon

    # All lines have trailing semicolon
    _check(
        txt=dedent(
            """\
        chr1 AUGUSTUS gene 68330 73621 1 - . ID=g1903;
        chr1 AUGUSTUS mRNA 68330 73621 1 - . ID=g1903.t1;Parent=g1903;
        chr1 Pfam protein_match 73372 73618 1 - . ID=g1903.t1.d1;Parent=g1903.t1;
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1903.t1.d1.1;Parent=g1903.t1.d1;
        """
        ),
        expected_keys=[
            ["ID"],
            ["ID", "Parent"],
            ["ID", "Parent"],
            ["ID", "Parent"],
        ],
        dialect_trailing_semicolon=True,
    )

    # First two lines have trailing semicolon. However, the heuristics of
    # dialect selection, which favor attributes with more values (assuming more
    # information), decides that this file does NOT have trailing semicolons.
    _check(
        txt=dedent(
            """\
        chr1 AUGUSTUS gene 68330 73621 1 - . ID=g1903;
        chr1 AUGUSTUS mRNA 68330 73621 1 - . ID=g1903.t1;Parent=g1903;
        chr1 Pfam protein_match 73372 73618 1 - . ID=g1903.t1.d1;Parent=g1903.t1
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1903.t1.d1.1;Parent=g1903.t1.d1
        """
        ),
        expected_keys=[
            ["ID", ""],
            ["ID", "Parent", ""],
            ["ID", "Parent"],
            ["ID", "Parent"],
        ],
        dialect_trailing_semicolon=False,
    )

    # APPARENTLY INCONSISTENT: The only thing difference here is that the
    # Parent attribute has been removed, otherwise matches above (first two
    # have trailing semicolon). But now there are no empty keys.
    #
    # This is expected behavior, because there are no attributes with more keys
    # as above to give higher weight, and to break the tie between with and
    # without trailing semicolon, falls back to first dialect observed.
    _check(
        txt=dedent(
            """\
        chr1 AUGUSTUS gene 68330 73621 1 - . ID=g1903;
        chr1 AUGUSTUS mRNA 68330 73621 1 - . ID=g1903.t1;
        chr1 Pfam protein_match 73372 73618 1 - . ID=g1903.t1.d1
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1903.t1.d1.1
        """
        ),
        expected_keys=[["ID"], ["ID"], ["ID"], ["ID"]],
        dialect_trailing_semicolon=True,
    )

    # We can convince the heuristics to think there should be NO trailing
    # semicolon by giving one more line as evidence. Only difference is from
    # above is the last line.
    _check(
        txt=dedent(
            """\
        chr1 AUGUSTUS gene 68330 73621 1 - . ID=g1903;
        chr1 AUGUSTUS mRNA 68330 73621 1 - . ID=g1903.t1;
        chr1 Pfam protein_match 73372 73618 1 - . ID=g1903.t1.d1
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1903.t1.d1.1
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1904.t1.d1.1
        """
        ),
        expected_keys=[
            ["ID", ""],
            ["ID", ""],
            ["ID"],
            ["ID"],
            ["ID"],
        ],
        dialect_trailing_semicolon=False,
    )

    # Again seems inconsistent at first, but heuristics break ties by
    # preferring first dialect, which here is no trailing semicolon.
    _check(
        txt=dedent(
            """\
        chr1 AUGUSTUS gene 68330 73621 1 - . ID=g1903
        chr1 AUGUSTUS mRNA 68330 73621 1 - . ID=g1903.t1
        chr1 Pfam protein_match 73372 73618 1 - . ID=g1903.t1.d1;
        chr1 Pfam protein_hmm_match 73372 73618 1 - . ID=g1903.t1.d1.1;
        """
        ),
        expected_keys=[["ID"], ["ID"], ["ID", ""], ["ID", ""]],
        dialect_trailing_semicolon=False,
    )


def test_issue_213(tmp_path):
    # GFF header directives seem to be not parsed when building a db from
    # a file, even though it seems to work fine from a string.
    data = dedent(
        """
    ##gff-version 3
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    """
    )

    # Ensure directives are parsed from DataIterator
    it = gffutils.iterators.DataIterator(data, from_string=True)
    assert it.directives == ["gff-version 3"]

    # Ensure they're parsed into the db from a string
    db = gffutils.create_db(data, dbfn=":memory:", from_string=True, verbose=False)
    assert db.directives == ["gff-version 3"], db.directives

    tmp = tmp_path / "issue_213.gff3"
    with open(tmp, "w") as fout:
        fout.write(data + "\n")

    # Ensure they're parsed into the db from a file path for both str/Path.
    for input_path in (str(tmp), tmp):
        db = gffutils.create_db(input_path, ":memory:")
        assert db.directives == ["gff-version 3"], db.directives
        assert len(db.directives) == 1

    # Ensure they're parsed into the db for all str/Path input-output
    # combinations when both source and destination are file-backed.
    for input_path, output_path in (
        (str(tmp), str(tmp_path / "issue_213_str_str.db")),
        (str(tmp), tmp_path / "issue_213_str_path.db"),
        (tmp, str(tmp_path / "issue_213_path_str.db")),
        (tmp, tmp_path / "issue_213_path_path.db"),
    ):
        db = gffutils.create_db(input_path, dbfn=output_path, force=True)
        assert db.directives == ["gff-version 3"], db.directives
        assert len(db.directives) == 1



def test_pathlike_inputs(tmp_path):
    """
    Ensure various functions work with Path and str.
    """
    gff = Path(gffutils.example_filename("FBgn0031208.gff"))
    fasta = Path(gffutils.example_filename("dm6-chr2L.fa"))
    db_path = tmp_path / "pathlike.db"
    out_path = tmp_path / "pathlike.gff3"
    staged_gff = tmp_path / "pathlike-input.gff3"
    staged_gff.write_text(gff.read_text())
    staged_gff_db = Path("%s.%s" % (staged_gff, ".db"))

    db = gffutils.create_db(gff, db_path, force=True)
    assert db.dbfn == os.fspath(db_path)

    reopened = gffutils.FeatureDB(db_path)
    reopened.delete([], make_backup=True)
    assert (tmp_path / "pathlike.db.bak").exists()

    writer = GFFWriter(out_path)
    writer.write_rec(next(reopened.all_features()))
    writer.close()
    assert out_path.exists()

    assert helpers.is_gff_db(db_path)
    gffutils.create_db(staged_gff, staged_gff_db, force=True)
    assert helpers.get_gff_db(staged_gff) == os.fspath(staged_gff_db)

    seq = reopened["FBgn0031208"].sequence(fasta)
    expected_seq = reopened["FBgn0031208"].sequence(os.fspath(fasta))
    assert seq == expected_seq

def test_issue_212():


    data = dedent(
        """
    NC_000964.3	RefSeq	CDS	410	1747	.	+	0	gene_id "BSU_00010"; transcript_id "unassigned_transcript_1"; db_xref "EnsemblGenomes-Gn:BSU00010"; db_xref "EnsemblGenomes-Tr:CAB11777"; db_xref "GOA:P05648"; db_xref "InterPro:IPR001957"; db_xref "InterPro:IPR003593"; db_xref "InterPro:IPR010921"; db_xref "InterPro:IPR013159"; db_xref "InterPro:IPR013317"; db_xref "InterPro:IPR018312"; db_xref "InterPro:IPR020591"; db_xref "InterPro:IPR024633"; db_xref "InterPro:IPR027417"; db_xref "PDB:4TPS"; db_xref "SubtiList:BG10065"; db_xref "UniProtKB/Swiss-Prot:P05648"; db_xref "GenBank:NP_387882.1"; db_xref "GeneID:939978"; experiment "publication(s) with functional evidences, PMID:2167836, 2846289, 12682299, 16120674, 1779750, 28166228"; gbkey "CDS"; gene "dnaA"; locus_tag "BSU_00010"; note "Evidence 1a: Function from experimental evidences in the studied strain; PubMedId: 2167836, 2846289, 12682299, 16120674, 1779750, 28166228; Product type f : factor"; product "chromosomal replication initiator informational ATPase"; protein_id "NP_387882.1"; transl_table "11"; exon_number "1";
        """
    )
    inferred_dialect = gffutils.helpers.infer_dialect(data.split('\t')[-1])
    assert inferred_dialect["semicolon in quotes"]

    f = next(iter(gffutils.DataIterator(data, from_string=True, dialect=inferred_dialect)))
    assert f.dialect["semicolon in quotes"]
    assert f.attributes["note"] == ["Evidence 1a: Function from experimental evidences in the studied strain; PubMedId: 2167836, 2846289, 12682299, 16120674, 1779750, 28166228; Product type f : factor"]



import tempfile
from gffutils import parser, create, feature, iterators, constants, helpers, exceptions
from gffutils import example_filename, create_db
from . import attr_test_cases
from textwrap import dedent

import pytest

TEST_FILENAMES = [
    example_filename(i)
    for i in [
        "c_elegans_WS199_ann_gff.txt",
        "ensembl_gtf.txt",
        "hybrid1.gff3",
        "ncbi_gff3.txt",
        "c_elegans_WS199_dna_shortened.fa",
        "F3-unique-3.v2.gff",
        "jgi_gff2.txt",
        "wormbase_gff2_alt.txt",
        "c_elegans_WS199_shortened_gff.txt",
        "glimmer_nokeyval.gff3",
        "mouse_extra_comma.gff3",
        "wormbase_gff2.txt",
    ]
]


def test_directives():
    data = dedent(
        """
    ##directive1 example
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    .	.	.	.	.	.	.	.
    """
    )

    it = iterators.DataIterator(data, from_string=True)
    assert it.directives == ["directive1 example"]

    db = create_db(data, dbfn=":memory:", from_string=True, verbose=False)
    assert db.directives == ["directive1 example"], db.directives


@pytest.mark.parametrize("item", attr_test_cases.attrs)
def test_attrs_OK(item):
    """
    Given an attribute string and a dictionary of what you expect, test the
    attribute splitting and reconstruction (invariant roundtrip).

    There are some corner cases for the roundtrip invariance that don't work
    (see attr_test_cases.py for details); `acceptable_reconstruction` handles
    those.
    """
    attr_str = item["str"]
    attr_dict = item["attrs"]
    acceptable_reconstruction = item["ok"]
    dialect_mods = item["dialect_mods"]

    result, inferred_dialect = parser._split_keyvals(attr_str)
    result = dict(result)
    assert result == attr_dict, result

    reconstructed = parser._reconstruct(result, inferred_dialect, keep_order=True)
    if acceptable_reconstruction:
        assert reconstructed == acceptable_reconstruction, reconstructed
    else:
        assert reconstructed == attr_str, reconstructed

    # Get the default dialect for comparison, and update it with any diffs
    # indicated by the test case
    default_dialect = constants.dialect.copy()
    default_dialect.update(dialect_mods)

    print(inferred_dialect)
    print(dialect_mods)
    print(attr_str)
    assert default_dialect == inferred_dialect



def parser_smoke_test():
    """
    Just confirm we can iterate completely through the test files....
    """
    # Don't show the warnings for tests
    import logging

    parser.logger.setLevel(logging.CRITICAL)
    for filename in TEST_FILENAMES:
        p = iterators._FileIterator(filename)
        for _ in p:
            continue


def test_empty_recontruct():
    """
    reconstructing attributes with incomplete information returns empty string
    """
    assert parser._reconstruct(None, constants.dialect) == ""
    with pytest.raises(exceptions.AttributeStringError):
        parser._reconstruct(dict(ID="asdf"), None)
    with pytest.raises(exceptions.AttributeStringError):
        parser._reconstruct(None, None)


def test_empty_split_keyvals():
    attrs, dialect = parser._split_keyvals(keyval_str=None)
    assert attrs == feature.dict_class()
    # assert dialect == constants.dialect
    assert dialect is None


def test_repeated_keys_conflict():
    """
    if dialect says repeated keys, but len(vals) > 1, then the keys are not
    actually repeated....
    """
    #
    # This is now only checked when infer_dialect is True -- and is disabled
    # when a dialect is provided
    #
    # dialect = constants.dialect.copy()
    # dialect['repeated keys'] = True
    # assert_raises(exceptions.AttributeStringError, parser._split_keyvals, "Parent=1,2,3", dialect)


def test_parser_from_string():
    # DEPRECATED
    #
    # _StringIterator has been removed and is instead handled by DataIterator
    # creating a temp file and returning a _FileIterator.
    pass


def test_valid_line_count():
    p = iterators._FileIterator(example_filename("ncbi_gff3.txt"))
    assert len(list(p)) == 17

    p = iterators._FileIterator(example_filename("hybrid1.gff3"))
    assert len(list(p)) == 6

    p = iterators._FileIterator(example_filename("FBgn0031208.gff"))
    assert len(list(p)) == 27


def test_inconsistent_dialect():
    """
    The second feature does not have a trailing semicolon (wormbase_gff2_alt is
    like this).  But since the first feature does, that's what the db's dialect
    is set to, which can cause errors when parsing attributes.
    """
    db = create.create_db(
        """
    chr1	.	gene	1	100	.	+	.	gene_id "gene1";
    chr1	.	mRNA	1	100	.	+	.	transcript_id "mRNA1"
    """,
        ":memory:",
        from_string=True,
    )
    items = list(db.all_features())
    print(items[0])
    # before, was ['"mRNA1'] -- note extra "
    assert items[1].attributes["transcript_id"] == ["mRNA1"], items[1].attributes[
        "transcript_id"
    ]


def test_attributes():
    s = "chr2L	FlyBase	mRNA	7529	9484	.	+	.	ID=FBtr0300690;Name=CG11023-RC;Parent=FBgn0031208;"
    f = feature.feature_from_line(s)
    f.keep_order = True
    assert str(f) == s, str(f)



# Portions copied over from BCBio.GFF.GFFParser

import re
import collections
from urllib import parse
from gffutils import constants
from gffutils.exceptions import AttributeStringError

import logging

formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)

# Regex for each separator that will be tested
quoted_semicolon_patterns = dict()

for sep in (" ; ", "; ", ";"):
    quoted_semicolon_patterns[sep] = re.compile(
        rf"""
            {re.escape(sep)}   # The separator we're considering (escaped for VERBOSE mode)
            (?=                # Positive lookahead: does remaining content match?
                (?:            # Start non-capturing group
                    [^"]       # Either: match any character that is NOT a quote
                    |          # OR
                    "[^"]*"    # Match a complete quoted string, specifically:
                               #   - opening quote ", followed by
                               #   - zero or more non-quote characters [^"]*
                               #   - followed by closing quote "
                )*             # Repeat the above pattern zero or more times
                $              # Until we reach the end of the string
            )                  # End of lookahead
        """,
        re.VERBOSE,
    )

# Encoding/decoding notes
# -----------------------
# From
# https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md#description-of-the-format:
#
#       GFF3 files are nine-column, tab-delimited, plain text files.
#       Literal use of tab, newline, carriage return, the percent (%) sign,
#       and control characters must be encoded using RFC 3986
#       Percent-Encoding; no other characters may be encoded. Backslash and
#       other ad-hoc escaping conventions that have been added to the GFF
#       format are not allowed. The file contents may include any character
#       in the set supported by the operating environment, although for
#       portability with other systems, use of Latin-1 or Unicode are
#       recommended.
#
#           tab (%09)
#           newline (%0A)
#           carriage return (%0D)
#           % percent (%25)
#           control characters (%00 through %1F, %7F)
#
#       In addition, the following characters have reserved meanings in
#       column 9 and must be escaped when used in other contexts:
#
#           ; semicolon (%3B)
#           = equals (%3D)
#           & ampersand (%26)
#           , comma (%2C)
#
#
# See also issue #98.
#
# Note that spaces are NOT supposed to be encoded. Yet some GFF files have
# spaces encoded anyway; in these cases round-trip invariance will not hold
# since the %20 will be decoded but not re-encoded.
_to_quote = "\n\t\r%;=&,"
_to_quote += "".join([chr(i) for i in range(32)])
_to_quote += chr(127)


# Caching idea from urllib.parse.Quoter, which uses a defaultdict for
# efficiency. Here we're sort of doing the reverse of the "reserved" idea used
# there.
class Quoter(collections.defaultdict):
    def __missing__(self, b):
        if b != "" and b in _to_quote:
            res = "%{:02X}".format(ord(b))
        else:
            res = b
        self[b] = res
        return res


quoter = Quoter()


def _split_keyvals(keyval_str, dialect=None):
    """
    Dialect detection requires partially parsing the attributes.
    """
    from gffutils import feature

    quals = feature.dict_class()

    if not keyval_str:
        return quals, dialect

    infer_dialect = False
    if dialect is None:
        infer_dialect = True
        dialect = {}

    # No known cases yet of different multival separator
    dialect["multival separator"] = ","

    # Detection for these dialect fields can work on the full attribute
    # string. Other detection needs to wait until we've further parsed the
    # attributes.
    if infer_dialect:
        dialect["trailing semicolon"] = keyval_str[-1] == ";"
        dialect["leading semicolon"] = keyval_str[0] == ";"
        semicolon_in_quotes = False
        sep = None
        for sep in (" ; ", "; ", ";"):
            parts = keyval_str.split(sep)
            if len(parts) > 1:
                # If naive split differs from more expensive regex, we infer there was
                # a semicolon within quoted value and we'll have to use the expensive
                # method later
                parts_regex = re.split(quoted_semicolon_patterns[sep], keyval_str)
                if parts != parts_regex:
                    semicolon_in_quotes = True
                break
        dialect["semicolon in quotes"] = semicolon_in_quotes
        dialect["field separator"] = sep

    if dialect["trailing semicolon"]:
        keyval_str = keyval_str.rstrip(";")

    if dialect["leading semicolon"]:
        keyval_str = keyval_str.lstrip(";")

    if dialect["semicolon in quotes"]:
        parts = re.split(
            quoted_semicolon_patterns[dialect["field separator"]], keyval_str
        )
    else:
        parts = keyval_str.split(dialect["field separator"])

    # The next stage of dialect inference works on the 'parts' -- unsplit
    # keyval pairs -- like:
    #
    #    parts = ["ID=001", "Name=gene1"]
    #
    # or
    #
    #    parts = ["gene_id ENSG001", "gene_biotype protein_coding"]
    #
    if infer_dialect:
        dialect["fmt"] = "gff3"

        # Note: so far, have not found cases where we need to check more than
        # the first item
        if "=" in parts[0]:
            dialect["fmt"] = "gff3"
            dialect["keyval separator"] = "="
        else:
            dialect["fmt"] = "gtf"
            dialect["keyval separator"] = " "

    # Now we split
    #
    #    parts = ["ID=001", "Name=gene1"]
    #
    # into
    #
    #   key_val_tuples = [("ID", "001"), ("Name", "gene1")]
    #
    # in a dialect-dependent manner.
    kvsep = dialect["keyval separator"]
    key_val_tuples = [p.split(kvsep) for p in parts]

    # With the split keys we can detect whether any are repeated
    if infer_dialect:
        keys = [i[0] for i in key_val_tuples]
        dialect["repeated keys"] = len(keys) != len(set(keys))

        # For dialect detection, this will help figure out if there is
        # inconsistent quoting across values. It will only be used in the loop
        # below if infer_dialect is True
        quoted_values = []

    # Now work splitting the keys if needed.
    for i in key_val_tuples:

        if len(i) == 2:
            # Easy, on-spec case
            key, val = i

        elif len(i) == 1:
            # By convention, no value becomes an empty string, e.g. when done
            # parsing,
            #
            #   "ID=001;is_gene;"
            #
            # will end up as:
            #
            #   {"ID": "001", "is_gene": ""}
            key = i[0]
            val = ""

        else:
            # Multiple *spaces* within quoted values are joined back together
            # without requiring a regex, in contrast to when there's *field*
            # separator like a semicolon in the values.
            #
            # That is:
            #
            #   attributes = 'gene_description "an important gene"; gene_id "g001"'
            #
            # when split on spaces, becomes
            #
            #   key_val_tuples = [("gene_description", "an", "important", "gene"), ("gene_id", "g001")]
            #
            # so here when we only keep the first token as a key, that first
            # key/val pair will become:
            #
            #   {
            #     "gene_description": ["an important gene"],
            #     "gene_id": ["g001"],
            #   }
            #
            # Another pathological case, this time for GFF3:
            #
            #   Alias=SGN-M1347;ID=T0028;Note=marker name(s): T0028 SGN-M1347 |identity=99.58|escore=2e-126
            #
            # will become the following:
            #
            #   {
            #     "Alias": ["SGN-M1347"],
            #     "ID": ["T0028"],
            #     "Note": ["marker name(s): T0028 SGN-M1347 |identity=99.58|escore=2e-126"],
            #   }
            #
            key = i[0]
            val = kvsep.join(i[1:])

        # By convention all values are lists, even if there's only one value
        # (or even no values)
        if key not in quals:
            quals[key] = []

        # This will run on every value, accumulating in quoted_values to check
        # later for consistency
        if infer_dialect:
            quoted = len(val) > 0 and val[0] == '"' and val[-1] == '"'
            quoted_values.append(quoted)
            dialect["quoted GFF2 values"] = quoted

        if dialect["quoted GFF2 values"] and val:
            val = val.strip('"')

        if val:
            # For repeated keys dialect, don't split on an internal comma. That is,
            #
            #   attributes = 'db_xref="g01,g02"; db_xref="XYZ"'
            #
            # becomes:
            #
            #   {
            #     "db_xref": ["g01,g02", "XYZ"]
            #    }
            #
            if dialect.get("repeated keys"):
                quals[key].append(val)

            # Otherwise, split but only if it's a comma without a space. So:
            #
            #    attributes = 'db_xref="g01,g02"'
            #
            # becomes
            #    {
            #      "db_xref": ["g01", "g02"]
            #    }
            # but
            #
            #    attributes = 'description="kinase, subunit 1"'
            #                                      ^ note the space here
            # becomes
            #    {
            #      "description": ["kinase, subunit 1"]
            #    }
            #
            else:
                # E.g. the "kinase, subunit 1" example above
                if ", " in val:
                    quals[key].append(val)
                else:
                    quals[key].extend(val.split(","))

    # If there was inconsistent quoting, we fall back to "not quoted" so
    # as to avoid incorrectly stripping off first and last quotes.
    if infer_dialect and len(set(quoted_values)) > 1:
        # Prior behavior was to use whatever the first value used
        dialect["quoted GFF2 values"] = quoted_values[0]

        # Though there could be an argument for considering quotes in mixed
        # cases to be part of the string, though technically they should be
        # %-encoded if so.
        # dialect["quoted GFF2 values"] = False

    # Handle unquoting of %-encoded values
    if not constants.ignore_url_escape_characters and dialect["fmt"] == "gff3":
        for key, vals in quals.items():
            unquoted = [parse.unquote(v) for v in vals]
            quals[key] = unquoted

    # Now that we're not supporting old Python versions we can rely on dict
    # insertion order
    if infer_dialect:
        dialect["order"] = list(quals.keys())

    return quals, dialect


def _reconstruct(keyvals, dialect, keep_order=False, sort_attribute_values=False):
    """
    Reconstructs the original attributes string according to the dialect.

    Parameters
    ==========
    keyvals : dict
        Attributes from a GFF/GTF feature

    dialect : dict
        Dialect containing info on how to reconstruct a string version of the
        attributes

    keep_order : bool
        If True, then perform sorting of attribute keys to ensure they are in
        the same order as those provided in the original file.  Default is
        False, which saves time especially on large data sets.

    sort_attribute_values : bool
        If True, then sort values to ensure they will always be in the same
        order.  Mostly only useful for testing; default is False.
    """
    if not dialect:
        raise AttributeStringError()
    if not keyvals:
        return ""
    parts = []

    # Re-encode when reconstructing attributes
    if constants.ignore_url_escape_characters or dialect["fmt"] != "gff3":
        attributes = keyvals
    else:
        attributes = {}
        for k, v in keyvals.items():
            attributes[k] = []
            for i in v:
                attributes[k].append("".join([quoter[j] for j in i]))

    # May need to split multiple values into multiple key/val pairs
    if dialect["repeated keys"]:
        items = []
        for key, val in attributes.items():
            if len(val) > 1:
                for v in val:
                    items.append((key, [v]))
            else:
                items.append((key, val))
    else:
        items = list(attributes.items())

    def sort_key(x):
        # sort keys by their order in the dialect; anything not in there will
        # be in arbitrary order at the end.
        try:
            return dialect["order"].index(x[0])
        except ValueError:
            return 1e6

    if keep_order:
        items.sort(key=sort_key)

    for key, val in items:

        # Multival sep is usually a comma:
        if val:
            if sort_attribute_values:
                val = sorted(val)

            val_str = dialect["multival separator"].join(val)

            if val_str:

                # Surround with quotes if needed
                if dialect["quoted GFF2 values"]:
                    val_str = '"%s"' % val_str

                # Typically "=" for GFF3 or " " otherwise
                part = dialect["keyval separator"].join([key, val_str])
            else:
                part = key
        else:
            if dialect["fmt"] == "gtf":
                # By convention, GTF attributes with no value are reconstructed
                # with an empty string. E.g.:
                #   'gene_id "gene1"; is_gene;'
                #
                # becomes
                #
                #   {
                #     "gene_id": "gene1",
                #     "is_gene": ""
                #    }
                #
                # and is printed as:
                #
                #   'gene_id "gene1"; is_gene "";'
                part = dialect["keyval separator"].join([key, '""'])
            else:
                part = key
        parts.append(part)

    # Typically ";" or "; "
    parts_str = dialect["field separator"].join(parts)

    # Sometimes need to add this
    if dialect["trailing semicolon"]:
        parts_str += ";"

    return parts_str



import gffutils
from textwrap import dedent

TEST_FILENAMES = [
    gffutils.example_filename(i)
    for i in [
        "c_elegans_WS199_ann_gff.txt",
        "ensembl_gtf.txt",
        "hybrid1.gff3",
        "ncbi_gff3.txt",
        "c_elegans_WS199_dna_shortened.fa",
        "F3-unique-3.v2.gff",
        "jgi_gff2.txt",
        "wormbase_gff2_alt.txt",
        "c_elegans_WS199_shortened_gff.txt",
        "glimmer_nokeyval.gff3",
        "mouse_extra_comma.gff3",
        "wormbase_gff2.txt",
    ]
]


def parser_smoke_test():
    """
    Just confirm we can iterate completely through the test files....
    """
    # Don't show the warnings for tests
    import logging

    gffutils.parser.logger.setLevel(logging.CRITICAL)
    for filename in TEST_FILENAMES:
        p = gffutils.iterators._FileIterator(filename)
        for i in p:
            continue


def test_file_iterator():
    db = gffutils.create_db(gffutils.example_filename("hybrid1.gff3"), ":memory:")
    n = len(list(db.all_features()))
    assert n == 6, n


def test_feature_iterator():
    it = gffutils.DataIterator(gffutils.example_filename("hybrid1.gff3"))

    # A DataIterator of a _FileIterator should be a _FeatureIterator
    it = gffutils.DataIterator((i for i in it))
    assert isinstance(it, gffutils.iterators._FeatureIterator), it

    it = gffutils.DataIterator(it)

    db = gffutils.create_db(it, ":memory:")
    n = len(list(db.all_features()))
    assert n == 6, n


def test_update():
    db = gffutils.create_db(
        gffutils.example_filename("FBgn0031208.gtf"),
        ":memory:",
        disable_infer_genes=True,
    )
    f = gffutils.feature.feature_from_line(
        'chr2L	gffutils_derived	gene	11500	12500	.	-	.	gene_id "fake";', strict=False
    )
    db.update([f], disable_infer_genes=True)
    for i in db.all_features():
        print(i.id)
    assert f == db["fake"]


def test_string_iterator():
    gtfdata = dedent(
        """
    chr1	a	testing	1	10	.	+	.	gene_id "fake"; n "2";
    chr1	b	testing	1	10	.	+	.	gene_id "fake"; n "1";
    """
    )
    data = gffutils.iterators.DataIterator(gtfdata, from_string=True)
    n = len(list(data))
    assert n == 2, n

    db = gffutils.create_db(gtfdata, ":memory:", from_string=True)
    n = len(list(db.all_features()))
    assert n == 2, n



from gffutils import iterators
from gffutils import interface
from collections import Counter
import sys


def inspect(
    data,
    look_for=["featuretype", "chrom", "attribute_keys", "feature_count"],
    limit=None,
    verbose=True,
):
    """
    Inspect a GFF or GTF data source.

    This function is useful for figuring out the different featuretypes found
    in a file (for potential removal before creating a FeatureDB).

    Returns a dictionary with a key for each item in `look_for` and
    a corresponding value that is a dictionary of how many of each unique item
    were found.

    There will always be a `feature_count` key, indicating how many features
    were looked at (if `limit` is provided, then `feature_count` will be the
    same as `limit`).

    For example, if `look_for` is ['chrom', 'featuretype'], then the result
    will be a dictionary like::

        {
            'chrom': {
                'chr1': 500,
                'chr2': 435,
                'chr3': 200,
                ...
                ...
            }.

            'featuretype': {
                'gene': 150,
                'exon': 324,
                ...
            },

            'feature_count': 5000

        }


    Parameters
    ----------
    data : str, FeatureDB instance, or iterator of Features
        If `data` is a string, assume it's a GFF or GTF filename.  If it's
        a FeatureDB instance, then its `all_features()` method will be
        automatically called. Otherwise, assume it's an iterable of Feature
        objects.

    look_for : list
        List of things to keep track of. Options are:

            - any attribute of a Feature object, such as chrom, source, start,
              stop, strand.

            - "attribute_keys", which will look at all the individual
              attribute keys of each feature

    limit : int
        Number of features to look at.  Default is no limit.

    verbose : bool
        Report how many features have been processed.

    Returns
    -------
    dict
    """

    results = {}
    obj_attrs = []
    for i in look_for:
        if i not in ["attribute_keys", "feature_count"]:
            obj_attrs.append(i)
        results[i] = Counter()

    attr_keys = "attribute_keys" in look_for

    d = iterators.DataIterator(data)
    feature_count = 0
    for f in d:
        if verbose:
            sys.stderr.write("\r%s features inspected" % feature_count)
            sys.stderr.flush()

        for obj_attr in obj_attrs:
            results[obj_attr].update([getattr(f, obj_attr)])

        if attr_keys:
            results["attribute_keys"].update(f.attributes.keys())

        feature_count += 1
        if limit and feature_count == limit:
            break

    new_results = {}
    for k, v in results.items():
        new_results[k] = dict(v)

    new_results["feature_count"] = feature_count
    return new_results



from gffutils.create import create_db
from gffutils.interface import FeatureDB
from gffutils.feature import Feature
from gffutils.iterators import DataIterator
from gffutils.helpers import example_filename
from gffutils.exceptions import FeatureNotFoundError, DuplicateIDError
from gffutils.version import version as __version__

__all__ = [
    "__version__",
    "create_db",
    "FeatureDB",
    "Feature",
    "DataIterator",
    "example_filename",
    "FeatureNotFoundError",
    "DuplicateIDError",
]



from gffutils import example_filename
import gffutils
import gffutils.biopython_integration as bp
import pytest

# Skip tests entirely if BioPython not available
pytest.importorskip('Bio')


def test_roundtrip():
    """
    Feature -> SeqFeature -> Feature should be invariant.
    """
    db_fname = gffutils.example_filename("gff_example1.gff3")
    db = gffutils.create_db(db_fname, ":memory:")
    feature = db["ENSMUSG00000033845"]
    feature.keep_order = True
    dialect = feature.dialect
    s = bp.to_seqfeature(feature)
    assert s.location.start == feature.start - 1
    assert s.location.end == feature.stop
    assert s.id == feature.id
    f = bp.from_seqfeature(s, dialect=dialect, keep_order=True)
    assert feature == f



"""
This module provides different kinds of iterators, all wrapped by the
DataIterator class which should generally be the only one used in practice.

The BaseIterator class allows "peeking" `checklines` lines into the data --
even if it's a consumable iterator -- in order to figure out what the dialect
is and therefore decide whether the data is GFF or GTF format, which is
important for figuring out how to construct the database.

"""
import os
import tempfile
import itertools
from contextlib import contextmanager
from gffutils.feature import feature_from_line
from gffutils.interface import FeatureDB
from gffutils import helpers
from textwrap import dedent
from urllib.request import urlopen
from urllib import parse as urlparse


class Directive(object):
    def __init__(self, line):
        self.info = line


class _BaseIterator(object):
    def __init__(
        self,
        data,
        checklines=10,
        transform=None,
        force_dialect_check=False,
        dialect=None,
    ):
        """
        Base class for iterating over features.  In general, you should use
        DataIterator -- so see the docstring of class for argument
        descriptions.


        All subclasses -- _FileIterator, _URLIterator, _FeatureIterator,
        _StringIterator -- gain the following behavior:

            - self.current_item and self.current_item_number are set on every
              iteration.  This is very useful for debugging, or reporting to
              the user exactly what item or line number caused the issue.

            - transform a Feature before it gets yielded, filter out a Feature

            - auto-detect dialect by peeking `checklines` items into the
              iterator, and then re-reading those, applying the detected
              dialect.  If multiple dialects are found, use
              helpers._choose_dialect to figure out the best one.

            - keep track of directives

        """
        self.data = data
        self.checklines = checklines
        self.current_item = None
        self.current_item_number = None
        self.dialect = None
        self._observed_dialects = []
        self.directives = []
        self.transform = transform
        self.warnings = []

        if force_dialect_check and dialect is not None:
            raise ValueError(
                "force_dialect_check is True, but a dialect " "is provided"
            )
        if force_dialect_check:
            # In this case, self.dialect remains None.  When
            # parser._split_keyvals gets None as a dialect, it tries to infer
            # a dialect.
            self._iter = self._custom_iter()
        elif dialect is not None:
            self.dialect = dialect
        else:
            # Otherwise, check some lines to determine what the dialect should
            # be
            _peek = self.peek(checklines)
            self._peek = _peek
            self.dialect = helpers._choose_dialect(_peek)

    def _custom_iter(self):
        raise NotImplementedError("Must define in subclasses")

    def __iter__(self):
        for i in self._custom_iter():
            i.dialect = self.dialect
            if self.transform:
                i = self.transform(i)
                if i:
                    yield i
            else:
                yield i

    def _directive_handler(self, directive):
        self.directives.append(directive[2:])


class _FileIterator(_BaseIterator):
    """
    Subclass for iterating over features provided as a filename
    """

    def peek(self, n):
        initial = []
        for i, feature in enumerate(self._custom_iter()):
            initial.append(feature)
            if i == n:
                break
        return initial

    def open_function(self, data):
        data = os.path.expanduser(data)
        if data.endswith(".gz"):
            import gzip

            return gzip.open(data)
        return open(data)

    def _custom_iter(self):
        self.directives = []
        valid_lines = 0
        with self.open_function(self.data) as fh:
            for i, line in enumerate(fh):
                if isinstance(line, bytes):
                    line = line.decode("utf-8")
                line = line.rstrip("\n\r")
                self.current_item = line
                self.current_item_number = i

                if line == "##FASTA" or line.startswith(">"):
                    return

                if line.startswith("##"):
                    self._directive_handler(line)
                    continue

                if line.startswith(("#")) or len(line) == 0:
                    continue

                # (If we got here it should be a valid line)
                valid_lines += 1
                yield feature_from_line(line, dialect=self.dialect)


class _UrlIterator(_FileIterator):
    """
    Subclass for iterating over features provided as a URL
    """

    @contextmanager
    def open_function(self, data):
        response = urlopen(data)

        # ideas from
        # http://stackoverflow.com/a/17537107
        # https://rationalpie.wordpress.com/2010/06/02/\
        #               python-streaming-gzip-decompression/
        if data.endswith(".gz"):
            import zlib

            d = zlib.decompressobj(16 + zlib.MAX_WBITS)
            READ_BLOCK_SIZE = 1024

            def _iter():
                last_line = ""
                while True:
                    data = response.read(READ_BLOCK_SIZE)
                    if not data:
                        break
                    data = "".join((last_line, d.decompress(data).decode()))
                    lines = data.split("\n")
                    last_line = lines.pop()
                    for line in lines:
                        yield line + "\n"
                yield last_line

        else:

            def _iter():
                for line in response.readlines():
                    if not line:
                        break
                    yield line.decode() + "\n"

        try:
            yield _iter()
        finally:
            response.close()


class _FeatureIterator(_BaseIterator):
    """
    Subclass for iterating over features that are already in an iterator
    """

    def peek(self, n):
        initial = []
        for i, feature in enumerate(self.data):
            initial.append(feature)
            if i == n:
                break

        # If self.data is generator-like, we need to patch it back together.
        if hasattr(self.data, "__next__"):
            self.data = itertools.chain(initial, self.data)
        return initial

    def _custom_iter(self):
        for i, feature in enumerate(self.data):
            self.current_item = feature
            self.current_item_number = i
            yield feature


def is_url(url):
    """
    Check to see if a URL has a valid protocol.

    Parameters
    ----------
    url : str or unicode

    Returns
    -------
    True if `url` has a valid protocol False otherwise.
    """
    try:
        return urlparse.urlparse(url).scheme in set(urlparse.uses_netloc).difference(
            [""]
        )
    except:
        return False


def DataIterator(
    data,
    checklines=10,
    transform=None,
    force_dialect_check=False,
    from_string=False,
    **kwargs,
):
    """
    Iterate over features, no matter how they are provided.

    Parameters
    ----------
    data : str, iterable of Feature objs, FeatureDB
        `data` can be a string (filename, URL, or contents of a file, if
        from_string=True), any arbitrary iterable of features, or a FeatureDB
        (in which case its all_features() method will be called).

    checklines : int
        Number of lines to check in order to infer a dialect.

    transform : None or callable
        If not None, `transform` should accept a Feature object as its only
        argument and return either a (possibly modified) Feature object or
        a value that evaluates to False.  If the return value is False, the
        feature will be skipped.

    force_dialect_check : bool
        If True, check the dialect of every feature.  Thorough, but can be
        slow.

    from_string : bool
        If True, `data` should be interpreted as the contents of a file rather
        than the filename itself.

    dialect : None or dict
        Provide the dialect, which will override auto-detected dialects.  If
        provided, you should probably also use `force_dialect_check=False` and
        `checklines=0` but this is not enforced.
    """
    if isinstance(data, os.PathLike):
        data = os.fspath(data)

    if isinstance(data, _BaseIterator):
        return data

    _kwargs = dict(
        data=data,
        checklines=checklines,
        transform=transform,
        force_dialect_check=force_dialect_check,
        **kwargs,
    )
    if isinstance(data, str):
        if from_string:
            tmp = tempfile.NamedTemporaryFile(delete=False)
            data = dedent(data)
            if isinstance(data, str):
                data = data.encode("utf-8")
            tmp.write(data)
            tmp.close()
            _kwargs["data"] = tmp.name
            return _FileIterator(**_kwargs)
        else:
            if os.path.exists(data):
                return _FileIterator(**_kwargs)
            elif is_url(data):
                return _UrlIterator(**_kwargs)
            else:
                raise ValueError(
                    f"{data} cannot be found and does not " "appear to be a URL"
                )
    elif isinstance(data, FeatureDB):
        _kwargs["data"] = data.all_features()
        return _FeatureIterator(**_kwargs)

    else:
        return _FeatureIterator(**_kwargs)



import warnings

from gffutils.helpers import asinterval

try:
    from pybedtools.contrib.plotting import Track
except ImportError:
    Track = None


class Gene(object):
    def __init__(
        self,
        db,
        gene_id,
        transcripts=["mRNA"],
        utrs=["3'UTR", "5'UTR"],
        cds=["CDS"],
        ybase=0,
        **kwargs
    ):
        """
        Represents a gene, `gene_id`, as a collection of
        pybedtools.contrib.plotting.Track objects.

        This class is flexible in how transcripts/CDSs/UTRs are defined;
        for example, you will usually have to adjust what featuretypes to
        consider as UTRs based on your annotation.

        For example, GTF files will typically define "3'UTR" but FlyBase GFF
        files use "three_prime_UTR".  In the former case you'd specify
        `utrs=["3'UTR", "5'UTR"]` while in the latter,
        `utrs=["three_prime_UTR", "five_prime_UTR"]`.

        The full length of the transcript is represented by a thin line; UTRs
        are slightly thicker, and CDSs are thickest.

        `kwargs` are passed to pybedtools.contrib.plotting.Track.

        `ybase` sets the bottom edge of the bottom isoform (sorted so that it's
        the shortest)

        Use the `max_y` attribute if you need to know the upper extent of the
        plotted transcripts.  This is useful if you are plotting multiple genes
        on the same axes and need to know what to use for the next one's
        `ybase`.

        You can adjust the `heights` attribute to set how wide transcripts,
        UTRs, CDSs are.  Padding is essentially "full" minus the largest height
        (CDS, 0.9, by default).
        """
        if Track is None:
            warnings.warn("Please install pybedtools for plotting.")
            raise ImportError("pybedtools is required for gffutils.contrib.plotting")

        self.heights = {"transcript": 0.2, "utrs": 0.5, "cds": 0.9, "full": 1.0}
        self.kwargs = kwargs
        self._transcripts = []
        for transcript in db.children(gene_id, level=1):
            if transcripts is None or transcript.featuretype in transcripts:
                d = {}
                d["transcript"] = [transcript]
                d["utrs"] = []
                d["cds"] = []
                for child in db.children(transcript, level=1):
                    _utrs = []
                    if child.featuretype in utrs:
                        d["utrs"].append(child)
                    if child.featuretype in cds:
                        d["cds"].append(child)
                self._transcripts.append(d)

        self.tracks = []

        self.ybase = ybase
        for d in sorted(
            self._transcripts, reverse=True, key=lambda x: len(x["transcript"])
        ):
            self.tracks.append(self._make_track(d, "transcript"))
            self.tracks.append(self._make_track(d, "utrs"))
            self.tracks.append(self._make_track(d, "cds"))
            self.ybase += self.heights["full"]

        self.max_y = ybase + self.heights["full"]

    def add_to_ax(self, ax):
        """
        Add all the transcripts for this gene to axes `ax`
        """
        for track in self.tracks:
            ax.add_collection(track)

    def _make_track(self, d, cls):
        yheight = self.heights[cls]
        ybase = self.ybase + (self.heights["full"] - yheight) * 0.5
        return Track(
            (asinterval(i) for i in d[cls]), ybase=ybase, yheight=yheight, **self.kwargs
        )



import collections

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections
from gffutils import constants


# collections.MutableMapping is apparently the best way to provide dict-like
# interface (http://stackoverflow.com/a/3387975)
class Attributes(collectionsAbc.MutableMapping):
    def __init__(self, *args, **kwargs):
        """
        An Attributes object acts much like a dictionary.  However, values are
        always stored internally as lists, even if a single value is provided.

        Whether or not you get a list back depends on the
        `constants.always_return_list` setting, which can be set on-the-fly.
        If True, then one-item lists are returned.  This is best shown with an
        example:

        Set up an Attributes object:

            >>> attr = Attributes()

        Set the "Name" attribute with a string:

            >>> attr['Name'] = 'gene1'

        This is stored internally as a list, and by default, we'll get a list
        back:

            >>> assert attr['Name'] == ['gene1']

        The same thing happens if we set it with a list in the first place:

            >>> attr['Name'] = ['gene1']
            >>> assert attr['Name'] == ['gene1']

        Now, change the setting so that upon access, single-value lists are
        returned as the first item.

            >>> constants.always_return_list = False
            >>> assert attr['Name'] == 'gene1'

        Change it back again:

            >>> constants.always_return_list = True
            >>> assert attr['Name'] == ['gene1']

        """
        self._d = dict()
        self.update(*args, **kwargs)

    def __setitem__(self, k, v):
        if not isinstance(v, (list, tuple)):
            v = [v]
        self._d[k] = v

    def __getitem__(self, k):
        v = self._d[k]
        if constants.always_return_list:
            return v
        if isinstance(v, list) and len(v) == 1:
            v = v[0]
        return v

    def __delitem__(self, key):
        del self._d[key]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self._d)

    def keys(self):
        return self._d.keys()

    def values(self):
        return [self.__getitem__(k) for k in self.keys()]

    def items(self):
        r = []
        for k in self.keys():
            r.append((k, self.__getitem__(k)))
        return r

    def __str__(self):
        s = []
        for i in self.items():
            s.append("%s: %s" % i)
        return "\n".join(s)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


# Useful for profiling: which dictionary-like class to store attributes in.
# This is used in Feature below and in parser.py

dict_class = Attributes
